#include <vector>
 
using namespace std;
typedef unsigned long long uint64_t_;

void SUFFIX(matrixSubset)(REALIN* in_, int inputNumIndividuals, int inputNumSNPs, std::vector<size_t> individuals_idx, std::vector<int> snpIdxList, REALOUT* out);
